/*
 * Copyright 2010 Trustees of the University of Pennsylvania Licensed under the
 * Educational Community License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License. You may
 * obtain a copy of the License at
 *
 * http://www.osedu.org/licenses/ECL-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS"
 * BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */
package metridoc.utils

import org.apache.camel.Navigate
import org.apache.camel.builder.RouteBuilder
import org.apache.camel.component.seda.SedaEndpoint
import org.apache.camel.processor.aggregate.AggregateProcessor
import org.slf4j.LoggerFactory
import org.apache.camel.model.RouteDefinition
import org.apache.camel.Endpoint
import org.apache.camel.model.FromDefinition
import org.apache.camel.Route

/**
 * Created by IntelliJ IDEA.
 * User: tbarker
 * Date: 8/26/11
 * Time: 3:45 PM
 * To change this template use File | Settings | File Templates.
 */
class CamelUtils {
    static final log = LoggerFactory.getLogger(CamelUtils.class)
    static final long pause = 100

    static void waitTillDone(RouteBuilder builder) {
        boolean notDone = true


        while (notDone) {
            notDone = !done(builder)
            if (notDone) {
                Thread.sleep(pause)
            }
        }

        log.debug("RouteBuilder {} has completed", builder)
    }

    static boolean done(RouteBuilder builder) {

        def routes = builder.routeCollection.routes
        for (RouteDefinition route: routes) {
            if (!isDone(builder, route)) {
                return false
            }
        }

        return true
    }

    static boolean isDone(RouteBuilder builder, RouteDefinition routeDefinition) {
        def context = builder.context


        for (FromDefinition from: routeDefinition.inputs) {

            def uri = from.uri
            def endpoint = context.getEndpoint(uri)

            if(endpoint instanceof SedaEndpoint) {

                for(Endpoint loadedEndpoint : context.endpoints) {
                    if(loadedEndpoint instanceof SedaEndpoint) {
                        def loadedUri = loadedEndpoint.endpointUri.replaceFirst("://", ":")
                        log.debug("comparing seda endpoints {} and {} for completion algorithm", uri, loadedUri)
                        if(uri.startsWith(loadedUri)) {
                            if (!isDone(loadedEndpoint)) {
                                return false
                            }
                        }
                    }
                }
            }

            if (!isDone(endpoint)) {
                return false
            }

        }
        return true
    }

    static boolean isDone(Endpoint endpoint) {
        def context = endpoint.camelContext
        log.debug("checking if endpoint {} is done", endpoint)


        if (inFlightExchangesDetected(endpoint)) {
            logFoundItem("inflight", endpoint)
            return false
        }

        if(sedaExchangesDetected(endpoint)) {
            logFoundItem("seda", endpoint)
            return false
        }

        if (aggregationExchangesDetected(endpoint)) {
            logFoundItem("aggregation", endpoint)
            return false
        }

        return true
    }

    static boolean aggregationExchangesDetected(Endpoint endpoint) {
        def context = endpoint.camelContext
        for (Route route: context.getRoutes()) {

            if (route.endpoint == endpoint) {
                def navigate = route.navigate()
                if (containsAggregationItems(navigate)) {

                    return true
                }

                break;
            }
        }
    }

    static boolean inFlightExchangesDetected(Endpoint endpoint) {
        def context = endpoint.camelContext
        def inflight = context.inflightRepository
        int inflightCount = inflight.size(endpoint)
        return inflightCount > 0
    }

    static boolean sedaExchangesDetected(Endpoint endpoint) {
        if (endpoint instanceof SedaEndpoint) {
            def seda = endpoint as SedaEndpoint
            def exchanges = seda.exchanges
            return exchanges.size() > 0
        }
    }

    private static void logFoundItem(String type, Endpoint endpoint) {
        log.debug("Route {} is not done since it found 1 or more {} exchanges", endpoint, type)
    }

    private static boolean containsAggregationItems(Navigate navigator) {
        if (navigator.hasNext()) {
            def processors = navigator.next()
            boolean hasItems = false
            processors.each {processor ->
                if (processor instanceof AggregateProcessor) {
                    def agg = processor as AggregateProcessor
                    hasItems = agg.aggregationRepository.keys.size() > 0
                }

                if (!hasItems) {
                    if (processor instanceof Navigate) {
                        hasItems = containsAggregationItems(processor)
                    }
                }
            }

            return hasItems
        }
    }

    static void waitTillDone(RouteBuilder builder, boolean shutDownRoutes) {
        waitTillDone(builder)

        if (shutDownRoutes) {
            shutDownBuilder(builder)
        }
    }

    static void shutDownBuilder(RouteBuilder builder) {
        def context = builder.context
        builder.routeCollection.routes.each {RouteDefinition definition ->
            context.stopRoute(definition)
        }
    }
}
